"""Container task type model and manager."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from .base_task_type import (
    BaseTaskType,
    BaseTaskTypeManager,
)


class ContainerTaskType(BaseTaskType):
    """Model for container task types.

    Attributes:
        id (int): The ID of the task type.
        name (str): The name of the task type.
        description (str): The description of the task type.
        user (str): The user associated with the task type.
        datetime_created (:class:`datetime.datetime`): The datetime when
            the task type was created.
        command_to_run (str): The command to run inside the container to
            execute the task.
        environment_variables (list): The environment variables required
            on the host to execute the task.
        required_arguments (list): The argument names for the task type.
        required_arguments_default_values (dict): Default values for the
            tasks required arguments.
        logs_path (str): The path of the logs directory inside the
            container.
        results_path (str): The path of the results directory inside the
            container.
        container_image (str): The container name and tag. For example,
            ubuntu:14.04 for Docker; and docker://ubuntu:14:04 or
            shub://vsoch/hello-world for Singularity.
        container_type (str): The type of the container.
        manager (:class:`saltant.models.container_task_type.ContainerTaskTypeManager`):
            The task type manager which spawned this task type.
    """
    def __init__(
            self,
            id,
            name,
            description,
            user,
            datetime_created,
            command_to_run,
            environment_variables,
            required_arguments,
            required_arguments_default_values,
            logs_path,
            results_path,
            container_image,
            container_type,
            manager,):
        """Initialize a container task type.

        Args:
            id (int): The ID of the task type.
            name (str): The name of the task type.
            description (str): The description of the task type.
            user (str): The user associated with the task type.
            datetime_created (:class:`datetime.datetime`): The datetime
                when the task type was created.
            command_to_run (str): The command to run to execute the task.
            environment_variables (list): The environment variables
                required on the host to execute the task.
            required_arguments (list): The argument names for the task type.
            required_arguments_default_values (dict): Default values for
                the tasks required arguments.
            logs_path (str): The path of the logs directory inside the
                container.
            results_path (str): The path of the results directory inside
                the container.
            container_image (str): The container name and tag. For
                example, ubuntu:14.04 for Docker; and docker://ubuntu:14:04
                or shub://vsoch/hello-world for Singularity.
            container_type (str): The type of the container.
            manager (:class:`saltant.models.container_task_type.ContainerTaskTypeManager`):
                The task type manager which spawned this task type.
        """
        # Call the parent constructor
        super(ContainerTaskType, self).__init__(
            id=id,
            name=name,
            description=description,
            user=user,
            datetime_created=datetime_created,
            command_to_run=command_to_run,
            environment_variables=environment_variables,
            required_arguments=required_arguments,
            required_arguments_default_values=required_arguments_default_values,
            manager=manager,
        )

        # Add in the attributes unique to container task types
        self.logs_path = logs_path
        self.results_path = results_path
        self.container_image = container_image
        self.container_type = container_type

    def put(self):
        """Updates this task type on the saltant server.

        Returns:
            :class:`saltant.models.container_task_type.ContainerTaskType`:
                A task type model instance representing the task type
                just updated.
        """
        return self.manager.put(
            id=self.id,
            name=self.name,
            description=self.description,
            command_to_run=self.command_to_run,
            environment_variables=self.environment_variables,
            required_arguments=self.required_arguments,
            required_arguments_default_values=(
                self.required_arguments_default_values),
            logs_path=self.logs_path,
            results_path=self.results_path,
            container_image=self.container_image,
            container_type=self.container_type,
        )


class ContainerTaskTypeManager(BaseTaskTypeManager):
    """Manager for Container task types.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task types.
        detail_url (str): The URL format to get specific task types.
        model (:class:`saltant.models.container_task_type.ContainerTaskType`):
            The model of the task instance being used.
    """
    list_url = 'containertasktypes/'
    detail_url = 'containertasktypes/{id}/'
    model = ContainerTaskType

    def create(
            self,
            name,
            command_to_run,
            container_image,
            container_type,
            description="",
            logs_path="",
            results_path="",
            environment_variables=None,
            required_arguments=None,
            required_arguments_default_values=None,
            extra_data_to_post=None,):
        """Create a container task type.

        Args:
            name (str): The name of the task.
            command_to_run (str): The command to run to execute the task.
            container_image (str): The container name and tag. For
                example, ubuntu:14.04 for Docker; and docker://ubuntu:14:04
                or shub://vsoch/hello-world for Singularity.
            container_type (str): The type of the container.
            description (str, optional): The description of the task type.
            logs_path (str, optional): The path of the logs directory
                inside the container.
            results_path (str, optional): The path of the results
                directory inside the container.
            environment_variables (list, optional): The environment
                variables required on the host to execute the task.
            required_arguments (list, optional): The argument names for
                the task type.
            required_arguments_default_values (dict, optional): Default
                values for the tasks required arguments.
            extra_data_to_post (dict, optional): Extra key-value pairs
                to add to the request data. This is useful for
                subclasses which require extra parameters.

        Returns:
            :class:`saltant.models.container_task_type.ContainerTaskType`:
                A container task type model instance representing the
                task type just created.
        """
        # Add in extra data specific to container task types
        if extra_data_to_post is None:
            extra_data_to_post = {}

        extra_data_to_post.update({
            'container_image': container_image,
            'container_type': container_type,
            'logs_path': logs_path,
            'results_path': results_path,
        })

        # Call the parent create function
        return super(ContainerTaskTypeManager, self).create(
            name=name,
            command_to_run=command_to_run,
            description=description,
            environment_variables=environment_variables,
            required_arguments=required_arguments,
            required_arguments_default_values=required_arguments_default_values,
            extra_data_to_post=extra_data_to_post,)

    def put(self,
            id,
            name,
            description,
            command_to_run,
            environment_variables,
            required_arguments,
            required_arguments_default_values,
            logs_path,
            results_path,
            container_image,
            container_type,
            extra_data_to_put=None):
        """Updates a task type on the saltant server.

        Args:
            id (int): The ID of the task type.
            name (str): The name of the task type.
            description (str): The description of the task type.
            command_to_run (str): The command to run to execute the task.
            environment_variables (list): The environment variables
                required on the host to execute the task.
            required_arguments (list): The argument names for the task type.
            required_arguments_default_values (dict): Default values for
                the tasks required arguments.
            extra_data_to_put (dict, optional): Extra key-value pairs to
                add to the request data. This is useful for subclasses
                which require extra parameters.
            logs_path (str): The path of the logs directory inside the
                container.
            results_path (str): The path of the results directory inside
                the container.
            container_image (str): The container name and tag. For
                example, ubuntu:14.04 for Docker; and docker://ubuntu:14:04
                or shub://vsoch/hello-world for Singularity.
            container_type (str): The type of the container.
        """
        # Add in extra data specific to container task types
        if extra_data_to_put is None:
            extra_data_to_put = {}

        extra_data_to_put.update({
            'logs_path': logs_path,
            'results_path': results_path,
            'container_image': container_image,
            'container_type': container_type,
        })

        # Call the parent create function
        return super(ContainerTaskTypeManager, self).put(
            id=id,
            name=name,
            description=description,
            command_to_run=command_to_run,
            environment_variables=environment_variables,
            required_arguments=required_arguments,
            required_arguments_default_values=(
                required_arguments_default_values),
            extra_data_to_put=extra_data_to_put,
        )



"""Executable task type model and manager."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from .base_task_type import (
    BaseTaskType,
    BaseTaskTypeManager,
)


class ExecutableTaskType(BaseTaskType):
    """Model for executable task types.

    Attributes:
        id (int): The ID of the task type.
        name (str): The name of the task type.
        description (str): The description of the task type.
        user (str): The user associated with the task type.
        datetime_created (:class:`datetime.datetime`): The datetime when
            the task type was created.
        command_to_run (str): The command to run to execute the task.
        environment_variables (list): The environment variables required
            on the host to execute the task.
        required_arguments (list): The argument names for the task type.
        required_arguments_default_values (dict): Default values for the
            tasks required arguments.
        manager (:class:`saltant.models.executable_task_type.ExecutableTaskTypeManager`):
            The task type manager which spawned this task type instance.
    """
    pass


class ExecutableTaskTypeManager(BaseTaskTypeManager):
    """Manager for executable task types.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task types.
        detail_url (str): The URL format to get specific task types.
        model (:class:`saltant.models.executable_task_type.ExecutableTaskType`):
            The model of the task instance being used.
    """
    list_url = 'executabletasktypes/'
    detail_url = 'executabletasktypes/{id}/'
    model = ExecutableTaskType



"""Contains base classes for models and related classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from saltant.exceptions import BadHttpRequestError
from saltant.constants import (
    HTTP_200_OK,
)

class Model(object):
    """Base class for representing a model.

    Attributes:
        manager (:class:`saltant.models.resource.Manager`):
            The manager which spawned this model instance.
    """
    def __init__(self, manager):
        """Initialize the model.

        Args:
            manager (:class:`saltant.models.resource.Manager`):
                The manager which spawned this model instance.
        """
        self.manager = manager


class ModelManager(object):
    """Base class for a model manager.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list models.
        detail_url (str): The URL format to get specific models.
        model (:class:`saltant.models.resource.Model`): The model
            being used.
    """
    list_url = "NotImplemented"
    detail_url = "NotImplemented"
    model = Model

    def __init__(self, _client):
        """Save the client so we can make API calls in the manager.

        Args:
            _client (:class:`saltant.client.Client`): An
                authenticated saltant client.
        """
        self._client = _client

    def list(self, filters=None):
        """List model instances.

        Currently this gets *everything* and iterates through all
        possible pages in the API. This may be unsuitable for production
        environments with huge databases, so finer grained page support
        should likely be added at some point.

        Args:
            filters (dict, optional): API query filters to apply to the
                request. For example:

                .. code-block:: python

                    {'name__startswith': 'azure',
                     'user__in': [1, 2, 3, 4],}

                See saltant's API reference at
                https://mwiens91.github.io/saltant/ for each model's
                available filters.

        Returns:
            list:
                A list of :class:`saltant.models.resource.Model`
                subclass instances (for example, container task type
                model instances).
        """
        # Add in the page and page_size parameters to the filter, such
        # that our request gets *all* objects in the list. However,
        # don't do this if the user has explicitly included these
        # parameters in the filter.
        if not filters:
            filters = {}

        if 'page' not in filters:
            filters['page'] = 1

        if 'page_size' not in filters:
            # The below "magic number" is 2^63 - 1, which is the largest
            # number you can hold in a 64 bit integer. The main point
            # here is that we want to get everything in one page (unless
            # otherwise specified, of course).
            filters['page_size'] = 9223372036854775807

        # Form the request URL - first add in the query filters
        query_filter_sub_url = ''

        for idx, filter_param in enumerate(filters):
            # Prepend '?' or '&'
            if idx == 0:
                query_filter_sub_url += '?'
            else:
                query_filter_sub_url += '&'

            # Add in the query filter
            query_filter_sub_url += '{param}={val}'.format(
                param=filter_param,
                val=filters[filter_param],
            )

        # Stitch together all sub-urls
        request_url = (
            self._client.base_api_url
            + self.list_url
            + query_filter_sub_url)

        # Make the request
        response = self._client.session.get(request_url)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_200_OK,)

        # Return a list of model instances
        return self.response_data_to_model_instances_list(response.json())

    def get(self, id):
        """Get the model instance with a given id.

        Args:
            id (int or str): The primary identifier (e.g., pk or UUID)
                for the task instance to get.

        Returns:
            :class:`saltant.models.resource.Model`:
                A :class:`saltant.models.resource.Model` subclass
                instance representing the resource requested.
        """
        # Get the object
        request_url = (
            self._client.base_api_url
            + self.detail_url.format(id=id))

        response = self._client.session.get(request_url)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_200_OK,)

        # Return a model instance
        return self.response_data_to_model_instance(response.json())

    def response_data_to_model_instance(self, response_data):
        """Convert get response data to a model.

        Args:
            response_data (dict): The data from the request's response.

        Returns:
            :class:`saltant.models.resource.Model`:
                A :class:`saltant.models.resource.Model` subclass
                instance representing the resource given in the
                request's response data.
        """
        # Add in this manager to the data
        response_data['manager'] = self

        # Instantiate a model
        return self.model(**response_data)

    def response_data_to_model_instances_list(self, response_data):
        """Convert list response data to a list of models.

        Args:
            response_data (dict): The data from the request's response.

        Returns:
            list:
                A list of :class:`saltant.models.resource.Model`
                subclass instances.
        """
        return [self.response_data_to_model_instance(subdata)
                for subdata in response_data['results']]

    @staticmethod
    def validate_request_success(
            response_text,
            request_url,
            status_code,
            expected_status_code,):
        """Validates that a request was successful.

        Args:
            response_text (str): The response body of the request.
            request_url (str): The URL the request was made at.
            status_code (int): The status code of the response.
            expected_status_code (int): The expected status code of the
                response.

        Raises:
            :class:`saltant.exceptions.BadHttpRequestError`: The HTTP
                request failed.
        """
        try:
            assert status_code == expected_status_code
        except AssertionError:
            msg = ("Request to {url} failed with status {status_code}:\n"
                   "The reponse from the request was as follows:\n\n"
                   "{content}").format(
                       url=request_url,
                       status_code=status_code,
                       content=response_text,)
            raise BadHttpRequestError(msg)



"""Base class for task type models."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import json
import dateutil.parser
from saltant.constants import (
    HTTP_200_OK,
    HTTP_201_CREATED,
)
from .resource import Model, ModelManager


class BaseTaskType(Model):
    """Base model for a task type.

    Attributes:
        id (int): The ID of the task type.
        name (str): The name of the task type.
        description (str): The description of the task type.
        user (str): The user associated with the task type.
        datetime_created (:class:`datetime.datetime`): The datetime when
            the task type was created.
        command_to_run (str): The command to run to execute the task.
        environment_variables (list): The environment variables required
            on the host to execute the task.
        required_arguments (list): The argument names for the task type.
        required_arguments_default_values (dict): Default values for the
            tasks required arguments.
        manager (:class:`saltant.models.base_task_type.BaseTaskTypeManager`):
            The task type manager which spawned this task type.
    """
    def __init__(
            self,
            id,
            name,
            description,
            user,
            datetime_created,
            command_to_run,
            environment_variables,
            required_arguments,
            required_arguments_default_values,
            manager,):
        """Initialize a task type.

        Args:
            id (int): The ID of the task type.
            name (str): The name of the task type.
            description (str): The description of the task type.
            user (str): The user associated with the task type.
            datetime_created (:class:`datetime.datetime`): The datetime
                when the task type was created.
            command_to_run (str): The command to run to execute the task.
            environment_variables (list): The environment variables
                required on the host to execute the task.
            required_arguments (list): The argument names for the task type.
            required_arguments_default_values (dict): Default values for
                the tasks required arguments.
            manager (:class:`saltant.models.base_task_type.BaseTaskTypeManager`):
                The task type manager which spawned this task type.
        """
        # Call parent constructor
        super(BaseTaskType, self).__init__(manager)

        self.id = id
        self.name = name
        self.description = description
        self.user = user
        self.datetime_created = datetime_created
        self.command_to_run = command_to_run
        self.environment_variables = environment_variables
        self.required_arguments = required_arguments
        self.required_arguments_default_values = (
            required_arguments_default_values)

    def __str__(self):
        """String representation of the task type."""
        return "%s (%s)" % (self.name, self.user)

    def sync(self):
        """Sync this model with latest data on the saltant server.

        Note that in addition to returning the updated object, it also
        updates the existing object.

        Returns:
            :class:`saltant.models.base_task_type.BaseTaskType`:
                This task type instance after syncing.
        """
        self = self.manager.get(id=self.id)

        return self

    def put(self):
        """Updates this task type on the saltant server.

        Returns:
            :class:`saltant.models.base_task_type.BaseTaskType`:
                A task type model instance representing the task type
                just updated.
        """
        return self.manager.put(
            id=self.id,
            name=self.name,
            description=self.description,
            command_to_run=self.command_to_run,
            environment_variables=self.environment_variables,
            required_arguments=self.required_arguments,
            required_arguments_default_values=(
                self.required_arguments_default_values),
        )


class BaseTaskTypeManager(ModelManager):
    """Base manager for task types.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task types.
        detail_url (str): The URL format to get specific task types.
        model (:class:`saltant.models.resource.Model`): The model of the
            task type being used.
    """
    model = BaseTaskType

    def get(self, id=None, name=None):
        """Get a task type.

        Either the id xor the name of the task type must be specified.

        Args:
            id (int, optional): The id of the task type to get.
            name (str, optional): The name of the task type to get.

        Returns:
            :class:`saltant.models.base_task_type.BaseTaskType`:
                A task type model instance representing the task type
                requested.

        Raises:
            ValueError: Neither id nor name were set *or* both id and
                name were set.
        """
        # Validate arguments - use an xor
        if not ((id is None) ^ (name is None)):
            raise ValueError(
                "Either id or name must be set (but not both!)")

        # If it's just ID provided, call the parent function
        if id is not None:
            return super(BaseTaskTypeManager, self).get(id=id)

        # Try getting the task type by name
        return self.list(filters={"name": name})[0]

    def create(
            self,
            name,
            command_to_run,
            description="",
            environment_variables=None,
            required_arguments=None,
            required_arguments_default_values=None,
            extra_data_to_post=None,):
        """Create a task type.

        Args:
            name (str): The name of the task.
            command_to_run (str): The command to run to execute the task.
            description (str, optional): The description of the task type.
            environment_variables (list, optional): The environment
                variables required on the host to execute the task.
            required_arguments (list, optional): The argument names for
                the task type.
            required_arguments_default_values (dict, optional): Default
                values for the tasks required arguments.
            extra_data_to_post (dict, optional): Extra key-value pairs
                to add to the request data. This is useful for
                subclasses which require extra parameters.

        Returns:
            :class:`saltant.models.base_task_instance.BaseTaskType`:
                A task type model instance representing the task type
                just created.
        """
        # Set None for optional list and dicts to proper datatypes
        if environment_variables is None:
            environment_variables = []

        if required_arguments is None:
            required_arguments = []

        if required_arguments_default_values is None:
            required_arguments_default_values = {}

        # Create the object
        request_url = self._client.base_api_url + self.list_url
        data_to_post = {
            "name": name,
            "description": description,
            "command_to_run": command_to_run,
            "environment_variables": json.dumps(environment_variables),
            "required_arguments": json.dumps(required_arguments),
            "required_arguments_default_values":
                json.dumps(required_arguments_default_values),
        }

        # Add in extra data if any was passed in
        if extra_data_to_post is not None:
            data_to_post.update(extra_data_to_post)

        response = self._client.session.post(request_url, data=data_to_post)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_201_CREATED,)

        # Return a model instance representing the task type
        return self.response_data_to_model_instance(response.json())

    def put(self,
            id,
            name,
            description,
            command_to_run,
            environment_variables,
            required_arguments,
            required_arguments_default_values,
            extra_data_to_put=None):
        """Updates a task type on the saltant server.

        Args:
            id (int): The ID of the task type.
            name (str): The name of the task type.
            description (str): The description of the task type.
            command_to_run (str): The command to run to execute the task.
            environment_variables (list): The environment variables
                required on the host to execute the task.
            required_arguments (list): The argument names for the task type.
            required_arguments_default_values (dict): Default values for
                the tasks required arguments.
            extra_data_to_put (dict, optional): Extra key-value pairs to
                add to the request data. This is useful for subclasses
                which require extra parameters.

        Returns:
            :class:`saltant.models.base_task_type.BaseTaskType`:
                A :class:`saltant.models.base_task_type.BaseTaskType`
                subclass instance representing the task type just
                updated.
        """
        # Update the object
        request_url = (
            self._client.base_api_url
            + self.detail_url.format(id=id))
        data_to_put = {
            "name": name,
            "description": description,
            "command_to_run": command_to_run,
            "environment_variables": json.dumps(environment_variables),
            "required_arguments": json.dumps(required_arguments),
            "required_arguments_default_values": json.dumps(
                required_arguments_default_values),
        }

        # Add in extra data if any was passed in
        if extra_data_to_put is not None:
            data_to_put.update(extra_data_to_put)

        response = self._client.session.put(request_url, data=data_to_put)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_200_OK,)

        # Return a model instance representing the task instance
        return self.response_data_to_model_instance(response.json())

    def response_data_to_model_instance(self, response_data):
        """Convert response data to a task type model.

        Args:
            response_data (dict): The data from the request's response.

        Returns:
            :class:`saltant.models.base_task_type.BaseTaskType`:
                A model instance representing the task type from the
                reponse data.
        """
        # Coerce datetime strings into datetime objects
        response_data['datetime_created'] = (
            dateutil.parser.parse(response_data['datetime_created']))

        # Instantiate a model for the task instance
        return super(
            BaseTaskTypeManager,
            self).response_data_to_model_instance(response_data)



"""Classes for task queue model and manager."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from saltant.constants import (
    HTTP_200_OK,
    HTTP_201_CREATED,
)
from .resource import Model, ModelManager


class TaskQueue(Model):
    """Base model for a task queue.

    Attributes:
        id (int): The ID of the task queue.
        user (str): The user who created the task queue.
        name (str): The name of the task queue.
        description (str): The description of the task queue.
        private (bool): A Booleon signalling whether the queue can only
            be used by its associated user.
        active (bool): A Booleon signalling whether the queue is active.
        manager (:class:`saltant.models.task_queue.TaskQueueManager`):
            The task queue manager which spawned this task queue.
    """
    def __init__(self,
                 id,
                 user,
                 name,
                 description,
                 private,
                 active,
                 manager,):
        """Initialize a task queue.

        Args:
            id (int): The ID of the task queue.
            user (str): The user who created the task queue.
            name (str): The name of the task queue.
            description (str): The description of the task queue.
            private (bool): A Booleon signalling whether the queue can
                only be used by its associated user.
            active (bool): A Booleon signalling whether the queue is
                active.
            manager (:class:`saltant.models.task_queue.TaskQueueManager`):
                The task queue manager which spawned this task instance.
        """
        # Call the parent constructor
        super(TaskQueue, self).__init__(manager)

        # Add in task queue stuff
        self.id = id
        self.user = user
        self.name = name
        self.description = description
        self.private = private
        self.active = active

    def __str__(self):
        """String representation of the task queue."""
        return self.name

    def sync(self):
        """Sync this model with latest data on the saltant server.

        Note that in addition to returning the updated object, it also
        updates the existing object.

        Returns:
            :class:`saltant.models.task_queue.TaskQueue`:
                This task queue instance after syncing.
        """
        self = self.manager.get(id=self.id)

        return self

    def put(self):
        """Updates this task queue on the saltant server.

        Returns:
            :class:`saltant.models.task_queue.TaskQueue`:
                A task queue model instance representing the task queue
                just updated.
        """
        return self.manager.put(
            id=self.id,
            name=self.name,
            description=self.description,
            private=self.private,
            active=self.active)


class TaskQueueManager(ModelManager):
    """Manager for task queues.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task queues.
        detail_url (str): The URL format to get specific task queues.
        model (:class:`saltant.models.task_queue.TaskQueue`): The model
            of the task queue being used.
    """
    list_url = 'taskqueues/'
    detail_url = 'taskqueues/{id}/'
    model = TaskQueue

    def get(self, id=None, name=None):
        """Get a task queue.

        Either the id xor the name of the task type must be specified.

        Args:
            id (int, optional): The id of the task type to get.
            name (str, optional): The name of the task type to get.

        Returns:
            :class:`saltant.models.task_queue.TaskQueue`:
                A task queue model instance representing the task queue
                requested.

        Raises:
            ValueError: Neither id nor name were set *or* both id and
                name were set.
        """
        # Validate arguments - use an xor
        if not ((id is None) ^ (name is None)):
            raise ValueError(
                "Either id or name must be set (but not both!)")

        # If it's just ID provided, call the parent function
        if id is not None:
            return super(TaskQueueManager, self).get(id=id)

        # Try getting the task queue by name
        return self.list(filters={"name": name})[0]

    def create(self,
               name,
               description="",
               private=False,
               active=True,):
        """Create a task queue.

        Args:
            name (str): The name of the task queue.
            description (str, optional): A description of the task queue.
            private (bool, optional): A boolean specifying whether the
                queue is exclusive to its creator. Defaults to False.
            active (bool, optional): A boolean specifying whether the
                queue is active. Default to True.

        Returns:
            :class:`saltant.models.task_queue.TaskQueue`:
                A task queue model instance representing the task queue
                just created.
        """
        # Create the object
        request_url = self._client.base_api_url + self.list_url
        data_to_post = {
            "name": name,
            "description": description,
            "private": private,
            "active": active,}

        response = self._client.session.post(request_url, data=data_to_post)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_201_CREATED,)

        # Return a model instance representing the task instance
        return self.response_data_to_model_instance(response.json())

    def put(self, id, name, description, private, active):
        """Updates a task queue on the saltant server.

        Args:
            id (int): The ID of the task queue.
            name (str): The name of the task queue.
            description (str): The description of the task queue.
            private (bool): A Booleon signalling whether the queue can
                only be used by its associated user.
            active (bool): A Booleon signalling whether the queue is
                active.

        Returns:
            :class:`saltant.models.task_queue.TaskQueue`:
                A task queue model instance representing the task queue
                just updated.
        """
        # Update the object
        request_url = (
            self._client.base_api_url
            + self.detail_url.format(id=id))
        data_to_put = {
            "name": name,
            "description": description,
            "private": private,
            "active": active,}

        response = self._client.session.put(request_url, data=data_to_put)

        # Validate that the request was successful
        self.validate_request_success(
            response_text=response.text,
            request_url=request_url,
            status_code=response.status_code,
            expected_status_code=HTTP_200_OK,)

        # Return a model instance representing the task instance
        return self.response_data_to_model_instance(response.json())



"""Contains the saltant API client."""

# TODO(mwiens91): add support for JWT auth tokens

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import functools
import os
import requests
from saltant.constants import (
    DEFAULT_TIMEOUT_SECONDS,
    HTTP_200_OK,
)
from saltant.exceptions import (
    AuthenticationError,
    BadEnvironmentError,
)
from saltant.models.container_task_instance import (
    ContainerTaskInstanceManager,
)
from saltant.models.container_task_type import (
    ContainerTaskTypeManager,
)
from saltant.models.executable_task_instance import (
    ExecutableTaskInstanceManager,
)
from saltant.models.executable_task_type import (
    ExecutableTaskTypeManager,
)
from saltant.models.task_queue import (
    TaskQueueManager,
)
from saltant.models.user import (
    UserManager,
)


class Client:
    """API client for communicating with a saltant server.

    Example:

        >>> from saltant.client import Client
        >>> client = Client(
        ...     base_api_url='https://shahlabjobs.ca/api/',
        ...     auth_token='p0gch4mp101fy451do9uod1s1x9i4a')

    Attributes:
        base_api_url (str): The URL of the saltant API.
        session (:class:`requests.Session`): A session object to make
            requests with.
        container_task_instances (:class:`saltant.models.container_task_instance.ContainerTaskInstanceManager`):
            A manager for performing actions related to container task
            instances.
        container_task_types (:class:`saltant.models.container_task_type.ContainerTaskTypeManager`):
            A manager for performing actions related to container task
            types.
        executable_task_instances (:class:`saltant.models.executable_task_instance.ExecutableTaskInstanceManager`):
            A manager for performing actions related to executable task
            instances.
        executable_task_types (:class:`saltant.models.executable_task_type.ExecutableTaskTypeManager`):
            A manager for performing actions related to executable task
            types.
        task_queues (:class:`saltant.models.task_queues.TaskQueueManager`):
            A manager for performing actions related to task queues.
        users (:class:`saltant.models.user.UserManager`):
            A manager for performing actions related to users.
    """
    def __init__(
            self,
            base_api_url,
            auth_token,
            default_timeout=DEFAULT_TIMEOUT_SECONDS,
            test_if_authenticated=True):
        """Initialize the saltant API client.

        Args:
            base_api_url (str): The URL of the saltant API.
            auth_token (str): The registered user's authentication token.
            default_timeout (int, optional): The maximum number
                of seconds to wait for a request to complete. Defaults
                to 90 seconds.
            test_if_authenticated (bool, optional): A flag signalling
                whether to try making a read-only authenticated request
                when initializing the client. This is useful for
                ensuring a working connection and correct authentication
                details as soon as possible. Defaults to True.
        """
        # The base URL of the saltant API
        self.base_api_url = base_api_url

        # Start a requests session
        self.session = requests.Session()
        self.session.headers.update(
            {'Authorization': 'Token ' + auth_token})

        # Test that we're authorized
        if test_if_authenticated:
            self.test_authentication()

        # Record the default timeout we want
        self.session.request = functools.partial(
            self.session.request,
            timeout=default_timeout)

        # Add in model managers
        self.container_task_instances = (
            ContainerTaskInstanceManager(_client=self))
        self.container_task_types = ContainerTaskTypeManager(_client=self)
        self.executable_task_instances = (
            ExecutableTaskInstanceManager(_client=self))
        self.executable_task_types = ExecutableTaskTypeManager(_client=self)
        self.task_queues = TaskQueueManager(_client=self)
        self.users = UserManager(_client=self)

    def test_authentication(self):
        """Test that the client is authorized.

        This currently assumes that read-only operations require
        authentication, which is the intended authentication protocol
        for saltant servers.

        Raises:
            :class:`saltant.exceptions.AuthenticationError`: The
                authentication provided was invalid.
        """
        response = self.session.get(self.base_api_url + 'users/')

        try:
            assert response.status_code == HTTP_200_OK
        except AssertionError:
            raise AuthenticationError('Authentication invalid!')

    @classmethod
    def from_env(cls, default_timeout=DEFAULT_TIMEOUT_SECONDS):
        """Return a client configured from environment variables.

        Essentially copying this:
        https://github.com/docker/docker-py/blob/master/docker/client.py#L43.

        The environment variables looked for are the following:

        .. envvar:: SALTANT_API_URL

            The URL of the saltant API. For example,
            https://shahlabjobs.ca/api/.

        .. envvar:: SALTANT_AUTH_TOKEN

            The registered saltant user's authentication token.

        Example:

            >>> from saltant.client import from_env
            >>> client = from_env()

        Args:
            default_timeout (int, optional): The maximum number of
                seconds to wait for a request to complete. Defaults to
                90 seconds.

        Returns:
            :class:`Client`: A saltant API client object.

        Raises:
            :class:`saltant.exceptions.BadEnvironmentError`: The user
                has an incorrectly configured environment.
        """
        # Get variables from environment
        try:
            base_api_url = os.environ['SALTANT_API_URL']
        except KeyError:
            raise BadEnvironmentError("SALTANT_API_URL not defined!")

        try:
            # Try to get an auth token
            auth_token = os.environ['SALTANT_AUTH_TOKEN']
        except KeyError:
            raise BadEnvironmentError("SALTANT_AUTH_TOKEN not defined!")

        # Return the configured client
        return cls(
            base_api_url=base_api_url,
            auth_token=auth_token,
            default_timeout=default_timeout,)


# Allow convenient import access to environment-configured client
from_env = Client.from_env



"""Classes for user model and manager."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from .resource import Model, ModelManager


class User(Model):
    """Base model for a user.

    Attributes:
        username (str): The user's username.
        email (str): The user's email.
        manager (:class:`saltant.models.user.UserManager`):
            The manager which spawned this user instance.
    """
    def __init__(self, username, email, manager):
        """Initialize a user.

        Args:
            username (str): The user's username.
            email (str): The user's email.
            manager (:class:`saltant.models.user.UserManager`):
                The manager which spawned this user instance.
        """
        # Call parent constructor
        super(User, self).__init__(manager)

        # Add in user stuff
        self.username = username
        self.email = email

    def sync(self):
        """Sync this model with latest data on the saltant server.

        Note that in addition to returning the updated object, it also
        updates the existing object.

        Returns:
            :class:`saltant.models.user.User`:
                This user instance after syncing.
        """
        self = self.manager.get(username=self.username)

        return self

    def __str__(self):
        """String representation of the user."""
        return self.username


class UserManager(ModelManager):
    """Manager for task queues.

    Attributes:
        _client (:class:`saltant.client.Client`): An authenticated
            saltant client.
        list_url (str): The URL to list task queues.
        detail_url (str): The URL format to get specific task queues.
        model (:class:`saltant.models.user.User`): The model of the task
            queue being used.
    """
    list_url = 'users/'
    detail_url = 'users/{id}/'
    model = User

    def get(self, username):
        """Get a user.

        Args:
            username (str): The username of the user to get.

        Returns:
            :class:`saltant.models.user.User`:
                A user model instance representing the user requested.
        """
        # Call the parent function
        return super(UserManager, self).get(id=username)
